/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <pool.h>
#include <bio.h>
#include <json.h>
#include "jsonfs.h"
#include "axosoft.h"

static char *Pkgname = "axosoft";
static char *Defmount = "/n/axosoft";

typedef struct Aux Aux;
struct Aux {
	Node *node;

	char *buf;	/* write buffer */
	int len;	/* length of buffer in use */
	int siz;	/* buffer size allocated */
};

static struct {
	Sess *session;
	Node root;
	long cachetime;
} Params;

long clk;
char *Debug;

static void
dumptree(Node *n, int in)
{
	for (; n; n = n->next){
		print("%*s %-16q %s %s %s\n",
			in, "",
			n->dir.name,
			(n->walk)? "walk": "",
			(n->scan)? "scan": "",
			(n->flush)? "flush": "");

		if (n->child)
			dumptree(n->child, in+2);
	}
}

static void
dircp(Dir *d, Dir *s)
{
	*d = *s;
	if(d->uid)
		d->uid = estrdup9p(d->uid);
	else
		d->uid = estrdup9p("jukefs");

	if(d->gid)
		d->gid = estrdup9p(d->gid);
	else
		d->gid = estrdup9p("jukefs");

	if(d->muid)
		d->muid = estrdup9p(d->muid);
	else
		d->muid = estrdup9p("jukefs");
	d->name = estrdup9p(d->name);
}

static char *
kstrdup(char **pp, char *s)
{
	if(s == nil)
		return *pp;

	if(*pp)
		free(*pp);
	*pp = estrdup9p(s);
	return *pp;
}			

static uint
hash(char *s)
{
	uint h;
	uchar *p;

	h = 0;
	for(p = (uchar*)s; *p; p++)
		h = h*37 + *p;
	return h;
}

static Dir
protodir(Dir *parent, char *name, int mode, vlong length, 
	char *uid, char *gid, char *muid, long mtime, long atime)
{
	Dir d;
	long now = time(nil);

	if(mtime == 0){
		if(parent)
			mtime = parent->mtime;
		if(mtime == 0)
			mtime = now;
	}

	if(atime == 0){
		if(parent)
			atime = parent->atime;
		if(atime == 0)
			atime = now;
	}

	/* cannot inherit or default these */
	d.name = estrdup9p(name);
	d.length = length;
	d.mode = mode;

	d.qid.type = (mode & DMDIR)? QTDIR: QTFILE;
	d.qid.path = hash(name);
	d.qid.vers = now;
	if(uid)
		d.uid = estrdup9p(uid);
	else
		d.uid = nil;
	if(gid)
		d.gid = estrdup9p(gid);
	else
		d.gid = nil;
	if(muid)
		d.muid = estrdup9p(muid);
	else
		d.muid = nil;
	d.mtime = mtime;
	d.atime = atime;

	return d;
}


Node *
newdir(Node *n, char *uid, char *gid, char *muid, char *name)
{
	Node *child;
	Dir *proto;

	for(child = n->child; child; child = child->next)
		if(strcmp(child->dir.name, name) == 0)
			break;
	if(child){
		kstrdup(&child->dir.uid, uid);
		kstrdup(&child->dir.gid, gid);
		kstrdup(&child->dir.muid, muid);
		if(child->last != Permenant)
			child->last = time(nil);
		return child;
	}

	proto = nil;
	if(n && n->child)
		proto = &n->child->dir;

	child = emalloc9p(sizeof(Node));
	child->dir = protodir(proto, name, 0555|DMDIR, 0, uid, gid, muid, 0, 0);
	child->last = time(nil);
	child->next = n->child;
	n->child = child;
	child->parent = n;
	return child;
}

static Node *
mkfile(Node *n, char *name, vlong length, char *uid, char *gid, char *muid, long mtime, long atime, char *buf, char *url)
{
	Node *child;
	Dir *proto;

	if(n->last != Permenant)
		n->last = time(nil);

	for(child = n->child; child; child = child->next)
		if(strcmp(child->dir.name, name) == 0)
			break;
	if(child){
		kstrdup(&child->buf, buf);
		kstrdup(&child->url, url);
		kstrdup(&child->dir.uid, uid);
		kstrdup(&child->dir.gid, gid);
		kstrdup(&child->dir.muid, muid);
		child->dir.length = length;	
		if(mtime)
			child->dir.mtime = mtime;
		if(atime)
			child->dir.mtime = atime;
		if(child->last != Permenant)
			child->last = time(nil);
		return child;
	}

	proto = nil;
	if(n && n->child)
		proto = &n->child->dir;

	child = emalloc9p(sizeof(Node));
	child->dir = protodir(proto, name, 0444, length, uid, gid, muid, mtime, atime);
	child->last = time(nil);
	child->next = n->child;
	n->child = child;
	child->parent = n;
	child->buf = buf;
	child->url = url;
	return child;
}

Node *
newfile(Node *n, char *name, char *uid, char *gid, char *muid, long mtime, long atime, char *str)
{
	char *buf, *r, *w, last;

	last = 0;
	buf = emalloc9p(strlen(str)+1 +1);		/* +1 for \0 and +1 for extra \n */

	/* exorcise ghosts of operating systems past */
	r = str;
	w = buf;
	while(*r){
		if(r[0] == '\r' && r[1] == '\n')
			r++;
		last = *w++ = *r++;
	}
	if(w > buf && last != '\n')
		*w++ = '\n';
	*w = 0;
	return mkfile(n, name, strlen(buf), uid, gid, muid, mtime, atime, buf, nil);
}

void
writable(Node *n)
{
	n->dir.mode |= 0222;
}

Node *
newurl(Node *n, char *name, char *uid, char *gid, char *muid, long mtime, vlong length, char *url)
{
	return mkfile(n, name, length, uid, gid, muid, mtime, mtime, nil, estrdup9p(url));
}


static void
responderrstr(Req *r)
{
	char e[ERRMAX];

	*e = 0;
	rerrstr(e, sizeof(e));
	respond(r, e);
}

static void
fsattach(Req *r)
{
	Aux *a;
	char *spec = r->ifcall.aname;

	if(spec && *spec){
		respond(r, "invalid attach specifier");
		return;
	}

	a = emalloc9p(sizeof(Aux));
	r->fid->aux = a;
	a->node = Params.root.child;
	
	r->ofcall.qid = a->node->dir.qid;
	r->fid->qid = r->ofcall.qid;
	respond(r, nil);
}

static char*
fsclone(Fid *fid, Fid *nfid)
{
	Aux *a = fid->aux;
	Aux *na;

	na = emalloc9p(sizeof(Aux));
	na->node = a->node;
	nfid->aux = na;
	return nil;
}

static int
dirgen(int slot, Dir *d, void *aux)
{
	Node *n;
	Aux *a = aux;

	for(n = a->node->child; n; n = n->next)
		if(--slot < 0)
			break;
	if(n == nil)
		return -1;
	dircp(d, &n->dir);
	return 0;
}

static int
up2date(Node *n, ulong now)
{
	if(n->last == Permenant)
		return 1;
	if(now - n->last < Params.cachetime)
		return 1;
	return 0;
}

static char*
fswalk1(Fid *fid, char *name, Qid *qid)
{
	Node *n;
	int dbg;
	long now;
	Aux *a = fid->aux;

	now = time(nil);

	dbg = 0;
	if(Debug && strstr(Debug, "walk")){
		dbg = 1;
		fprint(2, "walk %p %p %q -> %q\n", a, a->node, a->node->dir.name, name);
	}

	if(strcmp(name, "..") == 0){
		fid->qid = a->node->dir.qid;
		*qid = a->node->dir.qid;
		if(Debug && strstr(Debug, "walk"))
			fprint(2, "	dot-dot success\n");
		return nil;
	}

	/*
	 * only use cached directory info if the cache timeout has not expired,
	 * or if it is a permenant entry
	 */
	for(n = a->node->child; n; n = n->next){
		if(strcmp(n->dir.name, name) == 0 && up2date(n, now)){
			fid->qid = n->dir.qid;
			*qid = n->dir.qid;
			a->node = n;
			if(dbg) fprint(2, "	existing, success\n");
			return nil;
		}
	}

	/*
	 * auto-create the directory if we have a walk function,
	 * failing that do a directory scan.
	 */
	if(a->node->walk){
		if(dbg) fprint(2, "	have a walk func\n");
		if((*a->node->walk)(a->node, Params.session, name) == 0){
			n = a->node->child;
			fid->qid = n->dir.qid;
			*qid = n->dir.qid;
			a->node = n;
			if(dbg) fprint(2, "	walkfunc, success\n");
			return nil;
		}
	}
	else
	if(a->node->scan){
		if(dbg) fprint(2, "	have a scan func\n");
		if((*a->node->scan)(a->node, Params.session) == -1)
			return "scan failed";
		for(n = a->node->child; n; n = n->next){
			if(strcmp(n->dir.name, name) == 0){
				fid->qid = n->dir.qid;
				*qid = n->dir.qid;
				a->node = n;
				if(dbg) fprint(2, "	scanfunc, success\n");
				return nil;
			}
		}
	}

	if(Debug && strstr(Debug, "walk"))
		fprint(2, " fail\n");
	return "not found";
}

static void
fsstat(Req *r)
{
	Aux *a = r->fid->aux;

	dircp(&r->d, &a->node->dir);
	respond(r, nil);
}

static void
fscreate(Req *r)
{
	Aux *a = r->fid->aux;

	if((a->node->dir.mode & 0222) == 0){
		respond(r, "dir is read only");
		return;
	}
	if((r->ifcall.perm & DMDIR) == 0 && a->node->mkdir){
		respond(r, "dir create only");
		return;
	}
	if((r->ifcall.perm & DMDIR) != 0 && !a->node->mkdir){
		respond(r, "file create only");
		return;
	}

	free(a->buf);
	a->len = 0;
	a->siz = 0;
	a->buf = nil;

	respond(r, nil);
}

static void
fsopen(Req *r)
{
	Aux *a = r->fid->aux;
	Node *n;
	long now;

	if (r->fid->qid.type & QTDIR){
		n = a->node;
		now = time(nil);
		if(n->scan && (! n->child || now - n->last >= Params.cachetime)){
			if((*n->scan)(n, Params.session) < 0){
				respond(r, "scan() failed");
				return;
			}
		}
		respond(r, nil);
		return;
	}
	if((r->ifcall.mode & OWRITE) == OWRITE){
		if((a->node->dir.mode & 0222) == 0){
			respond(r, "file read only");
			return;
		}
		free(a->buf);
		a->len = 0;
		a->siz = 0;
		a->buf = nil;
	}
	if((r->ifcall.mode & OREAD) == OREAD){
		if((a->node->dir.mode & 0444) == 0){
			respond(r, "file write only");
			return;
		}
		n = a->node;
		if(n->url){
			free(a->buf);
			n->buf = fetch(Params.session, &n->dir.length, n->url);
			if(n->buf == nil){
				respond(r, "fetch file failed");
				return;
			}
		}
		
	}
	respond(r, nil);
}

static void
fsread(Req *r)
{
	Aux *a = r->fid->aux;
	Node *n;
	vlong off;
	long len, got;

	if (r->fid->qid.type & QTDIR){
		dirread9p(r, dirgen, a);
		respond(r, nil);
		return;
	}

	n = a->node;
	off = r->ifcall.offset;
	len = r->ifcall.count;
	got = n->dir.length;

	if(off >= got){
		r->ofcall.count = 0;
		respond(r, nil);
		return;
	}

	if(off+len >= got)
		len = got - off;

	memmove(r->ofcall.data, n->buf+off, len);
	r->ofcall.count = len;
	respond(r, nil);
}

static void
fswrite(Req *r)
{
	Aux *a = r->fid->aux;
	vlong off;
	long len;

	off = r->ifcall.offset;
	len = r->ifcall.count;

	if(off+len+1 >= a->siz){		/* +1 to allow for terminating zero */
		a->siz = off+len+128;
		a->buf = erealloc9p(a->buf, a->siz);
		a->node->dir.length = off+len+1;
	}
	memmove(a->buf+off, r->ifcall.data, len);
	if(a->len < off+len){
		a->len = off+len;
		a->buf[a->len] = 0;
	}
	r->ofcall.count = len;
	respond(r, nil);
}

static void
fsdestroyfid(Fid *fid)
{
	Node *n;
	Aux *a = fid->aux;
	
	if(a == nil)
		return;
	n = a->node;

	if(n->flush && a->buf && a->len)
		if((*n->flush)(n, Params.session, a->buf, a->len) < 0)
			fprint(2, "%s: %s write failed, %r\n", argv0, n->dir.name);
	free(a->buf);
	free(a);
}
	
Srv fs =
{
	.destroyfid =	fsdestroyfid,
	.attach=	fsattach,
	.open=		fsopen,
	.create=	fscreate,
	.read=		fsread,
	.write=		fswrite,
	.stat=		fsstat,
	.clone= 	fsclone,
	.walk1= 	fswalk1,
};
static void
usage(void)
{
	fprint(2, "usage: %s [-a] [-c secs] [-w dir] [-m dir] [-s name] [url]\n", argv0);
	fprint(2, "  -a       don't authenticate\n");
	fprint(2, "  -c secs  filesystem cache time in secs (default 60)\n");
	fprint(2, "  -w dir   dir where webfs is mounted (default /mnt/web)\n");
	fprint(2, "  -m dir   dir where jira will be mounted (default /n/jira)\n");
	fprint(2, "  -s dir   name posted into /srv (none posted by default)\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	long tout;
	Sess sess, *s;
	char *mtpt, *svs;
	Node skyhook, *root, *ch;

	quotefmtinstall();
	fmtinstall('F', fcallfmt);
	fmtinstall('D', dirfmt);

	s = &sess;
	memset(s, 0, sizeof(Sess));

	svs = nil;
	mtpt = nil;
	tout = 60;
	s->webfs = "/mnt/web";
	ARGBEGIN{
	case 'D':
		Debug = EARGF(usage());
		break;
	case 'c':
		tout = atoi(EARGF(usage()));
		break;
	case 'w':
		s->webfs = EARGF(usage());
		break;
	case 'm':
		mtpt = EARGF(usage());
		break;
	case 's':
		svs = EARGF(usage());
		break;
	case 'M':
		mainmem->flags |= POOL_NOREUSE|POOL_PARANOIA|POOL_ANTAGONISM;
		break;
	default:
		usage();
	}ARGEND;

	if((s->url = getenv(Pkgname)) == nil)
		sysfatal("%s url not set\n", Pkgname);

	if(strncmp(s->url, "http://", 7) != 0 &&
	   strncmp(s->url, "https://", 8) != 0)
		sysfatal("%s - not a valid url\n", s->url);

	if(dialweb(s, nil) < 0)
		sysfatal("cannot get session to webfs\n");

	if(auth(s, Pkgname) < 0)
		sysfatal("quthentication failed, %r\n");

	memset(&skyhook, 0, sizeof(skyhook));
	root = newdir(&skyhook, nil, nil, nil, "");
	root->last = Permenant;

	ch = newdir(root, nil, nil, nil, "defects");
	ch->scan = defectscan;
	ch->walk = defectwalk;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "incidents");
	ch->scan = incidentscan;
	ch->walk = incidentwalk;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "features");
	ch->scan = featurescan;
	ch->walk = featurewalk;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "tasks");
	ch->scan = taskscan;
	ch->walk = taskwalk;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "users");
	ch->scan = alluserscan;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "local");
	ch->scan = localscan;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "projects");
	ch->scan = projectscan;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "teams");
	ch->scan = teamscan;
	ch->last = Permenant;

	ch = newdir(root, nil, nil, nil, "workflows");
	ch->scan = workflowscan;
	ch->last = Permenant;

	if(!mtpt && !svs)
		mtpt = Defmount;

	Params.session = s;
	Params.cachetime = tout;
	memcpy(&Params.root, &skyhook, sizeof(Node));

	postmountsrv(&fs, svs, mtpt, MREPL|MCREATE);

	exits(0);
}
