/* ©  Copyright Steve Simon, 2013 */
#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include <auth.h>
#include <fcall.h>
#include <thread.h>
#include <9p.h>
#include <json.h>
#include "jsonfs.h"
#include "axosoft.h"

void
warn(char *fmt, ...)
{
	va_list ap;

	fprint(2, "%s: ", argv0);
	va_start(ap, fmt);
	vfprint(2, fmt, ap);
	va_end(ap);
}

char *
strlwr(char *str)
{
	char *s;

	for(s = str; *s; s++)
		*s = tolower(*s);
	return str;
}

char *
strupr(char *str)
{
	char *s;

	for(s = str; *s; s++)
		*s = toupper(*s);
	return str;
}

static void
dmphdrs(Sess *s)
{
	Dir *d;
	int n, m, i, fd;
	char buf[1024];

	snprint(buf, sizeof(buf), "%s/%d", s->webfs, s->conid);
	if((fd = open(buf, OREAD)) == -1)
		return;
	n = dirreadall(fd, &d);
	close(fd);

	for(i = 0; i < n; i++){
		if(strcmp(d[i].name, "parsed") == 0)
			continue;
		snprint(buf, sizeof(buf), "%s/%d/%s", s->webfs, s->conid, d[i].name);
		fprint(2, "	%-32s ", buf);
		if((fd = open(buf, OREAD)) != -1)
			if((m = read(fd, buf, sizeof(buf)-1)) > 0)
				write(2, buf, m);
		fprint(2, "\n");
		close(fd);
	}
	free(d);
}


int
httphdr(Sess *s, char *field, char *buf, int len)
{
	int n, i, fd;
	char tmp[1024];

	if(buf && len > 0)
		*buf = 0;

	fd = -1;
	snprint(tmp, sizeof(tmp), "%s/%d/%s", s->webfs, s->conid, field);
	for(i = 0; i < 4; i++){
		if((fd = open(tmp, OREAD)) >= 0)
			break;

		sleep(10);
	}
	if(fd == -1){
//dmphdrs(s);
		close(fd);
		return -1;
	}

	n = read(fd, buf, len-1);
	close(fd);

	if(n < 0)
		return -1;
	buf[n] = 0;
	return 0;
}

Jvalue *
exch(Sess *s, char *post, char *fmt, ...)
{
	Jvalue *v;
	va_list ap;
	uvlong then;
	Biobuf *bp;
	int e, n, fd;
	char *url, *a[4], *tmp, buf[1024], err[ERRMAX];

	then = nsec();

	va_start(ap, fmt);
	url = vsmprint(fmt, ap);
	va_end(ap);

	/* set headers */
	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid);
	fd = open(buf, OWRITE);
	assert(fd != -1);

	if(post){
		fprint(fd, "request POST");
		fprint(fd, "contenttype application/json");
	}

	/* Cannot use fprint() as it has a 256 byte buffer and some urls are bigger than this! */
	tmp = smprint("url %s/%s", s->url, url);
	n = strlen(tmp);
	if(write(fd, tmp, n) != n)
		sysfatal("url write failed, %r\n");
	free(tmp);

	if(s->headers)
		fprint(fd, "headers %s\n", s->headers);

	close(fd);

	if(post){
		/* send post body */
		if(Debug && strstr(Debug, "url"))
			fprint(2, "post: %s\n", post);

		snprint(buf, sizeof buf, "%s/%d/postbody", s->webfs, s->conid);
		fd = open(buf, OWRITE);
		assert(fd != -1);
		fprint(fd, "%s", post);
		close(fd);
	}

	/* read reply */
	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid);
	bp = Bopen(buf, OREAD);
	if(bp == nil){
		rerrstr(err, sizeof(err));

		if(tokenize(err, a, nelem(a)) >= 2)
			if((e = atoi(a[1])) > 0)
				httperrstr(e);

		if(Debug && strstr(Debug, "url"))
			fprint(2, "%s/%q       %fsec    failed: %r\n", s->url, url, (double)(nsec()-then) / 1.0e9);
		free(url);
		return nil;
	}

	v = json_parse(bp, nil);
	Bterm(bp);

	if(Debug && strstr(Debug, "url"))
		fprint(2, "%s/%q       %fsec\n", s->url, url, (double)(nsec()-then) / 1.0e9);

	if(Debug && strstr(Debug, "json")){
		Biobuf out;
		Binit(&out, 2, OWRITE);
		json_print(&out, v);
		Bterm(&out);
	}
	free(url);
	return v;
}

char *
fetch(Sess *s, vlong *len, char *fmt, ...)
{
	va_list ap;
	int fd, got, siz, e, n;
	char *tmp, buf[1024], err[ERRMAX], *res, *a[8], *url;
	
	va_start(ap, fmt);
	url = vsmprint(fmt, ap);
	va_end(ap);

	if(Debug && strstr(Debug, "url"))
		fprint(2, "request: %s/%q\n", s->url, url);

	snprint(buf, sizeof buf, "%s/%d/ctl", s->webfs, s->conid);
	fd = open(buf, OWRITE);
	assert(fd != -1);

	/* Cannot use fprint() as it has a 256 byte buffer and some urls are bigger than this! */
	tmp = smprint("url %s/%s", s->url, url);
	n = strlen(tmp);
	if(write(fd, tmp, n) != n)
		sysfatal("url write failed, %r\n");
	free(tmp);

	if(s->headers)
		fprint(fd, "headers %s\n", s->headers);

	close(fd);
	free(url);

	snprint(buf, sizeof buf, "%s/%d/body", s->webfs, s->conid);
	fd = open(buf, OREAD);
	if(fd < 0){
		rerrstr(err, sizeof(err));
		if(tokenize(err, a, nelem(a)) >= 2)
			if((e = atoi(a[1])) > 0)
				httperrstr(e);
		return nil;
	}

	n = 0;
	got = 0;
	siz = 0;
	res = nil;
	do{
		got += n;
		if(siz - got < 4096){
			siz += 8192;
			res = erealloc9p(res, siz);
		}
	}while((n = read(fd, res+got, siz-got)) > 0);
	close(fd);
	*len = got;
	return res;
}

int
dialweb(Sess *s, char *relm)
{
	int n, fd;
	char buf[1024];

	snprint(buf, sizeof(buf), "%s/clone", s->webfs);
	if((fd = open(buf, ORDWR)) < 0)
		return -1;
	if((n = read(fd, buf, sizeof buf-1)) <= 0){
		close(fd);
		return -1;
	}
	buf[n] = '\0';
	if ((s->conid = atoi(buf)) <0){
		close(fd);
		return -1;
	}
	s->fd = fd;

	/*
	 * some servers need preauthentication as they can operate
	 * without authentication, but provide limited functionality
	 */
	if(relm){
		snprint(buf, sizeof(buf), "%s/ctl", s->webfs);
		if((fd = open(buf, ORDWR)) < 0)
			return -1;
		fprint(fd, "preauth %s %s", s->url, relm);
		close(fd);
	}

	return 0;
}

long
iso2sec(char *s)
{
	Tm tm;
	long t;
	char *td[4], *ymd[5], *hms[5], *tz;

	/* e.g.
	 * 2010-04-06T11:39:46Z
	 * 2011-11-27t00:00:00.000+0000
	 */
	memset(&tm, 0, sizeof(tm));

	if(s == nil || *s == 0)
		return -1;
	if(gettokens(s, td, nelem(td), "tT") != 2)
		return -1L;
	if(gettokens(td[0], ymd, nelem(ymd), "-") != 3)
		return -1L;
	if(gettokens(td[1], hms, nelem(hms), ":") != 2)

	tm.year = strtol(ymd[0], nil, 10) - 1900;
	tm.mon = strtol(ymd[1], nil, 10) -1;
	tm.mday = strtol(ymd[2], nil, 10);
	tm.hour = strtol(hms[0], nil, 10);
	tm.min = strtol(hms[1], nil, 10);
	tm.sec = strtol(hms[2], &tz, 10);

	if(*tz == '.')
		strtol(tz+1, &tz, 10);	/* skip fractional secconds */

	if(*tz == 'Z' || *tz == 'z')
		tm.tzoff = 0;
	else
	if(*tz == '+' || *tz == '-')
		tm.tzoff = strtol(tz, nil, 10);
	strcpy(tm.zone, "GMT");
	if((t = tm2sec(&tm)) < 0)
		return -1L;
	return t;
}

char *
basename(char *s)
{
	char *p;

	if((p = strrchr(s, '/')) == nil)
		return s;
	return p+1;
}

char *
getkey(char *attr, char *fmt, ...)
{
	Biobuf *bp;
	va_list ap;
	int nw, nf, i, j, len;
	char *line, *ret, *buf, *want[32], *found[32];

	len = strlen(attr);

	if((bp = Bopen("/mnt/factotum/ctl", OREAD)) == nil){
		fprint(2, "%s: /mnt/factotum/ctl cannot open, %r\n", argv0);
		return nil;
	}

	va_start(ap, fmt);
	buf = vsmprint(fmt, ap);
	va_end(ap);

	nw = tokenize(buf, want, nelem(want));
	
	ret = nil;
	while((line = Brdline(bp, '\n')) != nil){
		line[Blinelen(bp)-1] = 0;

		nf = tokenize(line, found, nelem(found));
		if(nf < nw)
			continue;

		for(i = 0; i < nw; i++){
			for(j = 0; j < nf; j++)
				if(strcmp(want[i], found[j]) == 0)
					break;
			if(j >= nf)				/* not this pair found */
				break;
		}
		if(i >= nw){				/* found them all */
			for(j = 0; j < nf; j++)
				if(strncmp(found[j], attr, len) == 0 && found[j][len] == '='){
					ret = strdup(found[j] +len +1);
					break;
				}
			break;
		}
	}

	free(buf);
	Bterm(bp);
	return ret;
}
